
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef JSON_COMPILATION
#define JSON_COMPILATION
#endif

#include "config.h"
#include "json-enum-types.h"

/* enumerations from "json-parser.h" */
#include "json-parser.h"
GType
json_parser_error_get_type (void)
{
  static gsize g_enum_type_id__volatile = 0;

  if (g_once_init_enter (&g_enum_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { JSON_PARSER_ERROR_PARSE, "JSON_PARSER_ERROR_PARSE", "parse" },
        { JSON_PARSER_ERROR_TRAILING_COMMA, "JSON_PARSER_ERROR_TRAILING_COMMA", "trailing-comma" },
        { JSON_PARSER_ERROR_MISSING_COMMA, "JSON_PARSER_ERROR_MISSING_COMMA", "missing-comma" },
        { JSON_PARSER_ERROR_MISSING_COLON, "JSON_PARSER_ERROR_MISSING_COLON", "missing-colon" },
        { JSON_PARSER_ERROR_INVALID_BAREWORD, "JSON_PARSER_ERROR_INVALID_BAREWORD", "invalid-bareword" },
        { JSON_PARSER_ERROR_EMPTY_MEMBER_NAME, "JSON_PARSER_ERROR_EMPTY_MEMBER_NAME", "empty-member-name" },
        { JSON_PARSER_ERROR_INVALID_DATA, "JSON_PARSER_ERROR_INVALID_DATA", "invalid-data" },
        { JSON_PARSER_ERROR_UNKNOWN, "JSON_PARSER_ERROR_UNKNOWN", "unknown" },
        { JSON_PARSER_ERROR_NESTING, "JSON_PARSER_ERROR_NESTING", "nesting" },
        { JSON_PARSER_ERROR_INVALID_STRUCTURE, "JSON_PARSER_ERROR_INVALID_STRUCTURE", "invalid-structure" },
        { JSON_PARSER_ERROR_INVALID_ASSIGNMENT, "JSON_PARSER_ERROR_INVALID_ASSIGNMENT", "invalid-assignment" },
        { 0, NULL, NULL }
      };

      GType g_enum_type_id =
        g_enum_register_static (g_intern_static_string ("JsonParserError"), values);

      g_once_init_leave (&g_enum_type_id__volatile, g_enum_type_id);
    }

  return g_enum_type_id__volatile;
}

/* enumerations from "json-path.h" */
#include "json-path.h"
GType
json_path_error_get_type (void)
{
  static gsize g_enum_type_id__volatile = 0;

  if (g_once_init_enter (&g_enum_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { JSON_PATH_ERROR_INVALID_QUERY, "JSON_PATH_ERROR_INVALID_QUERY", "query" },
        { 0, NULL, NULL }
      };

      GType g_enum_type_id =
        g_enum_register_static (g_intern_static_string ("JsonPathError"), values);

      g_once_init_leave (&g_enum_type_id__volatile, g_enum_type_id);
    }

  return g_enum_type_id__volatile;
}

/* enumerations from "json-reader.h" */
#include "json-reader.h"
GType
json_reader_error_get_type (void)
{
  static gsize g_enum_type_id__volatile = 0;

  if (g_once_init_enter (&g_enum_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { JSON_READER_ERROR_NO_ARRAY, "JSON_READER_ERROR_NO_ARRAY", "no-array" },
        { JSON_READER_ERROR_INVALID_INDEX, "JSON_READER_ERROR_INVALID_INDEX", "invalid-index" },
        { JSON_READER_ERROR_NO_OBJECT, "JSON_READER_ERROR_NO_OBJECT", "no-object" },
        { JSON_READER_ERROR_INVALID_MEMBER, "JSON_READER_ERROR_INVALID_MEMBER", "invalid-member" },
        { JSON_READER_ERROR_INVALID_NODE, "JSON_READER_ERROR_INVALID_NODE", "invalid-node" },
        { JSON_READER_ERROR_NO_VALUE, "JSON_READER_ERROR_NO_VALUE", "no-value" },
        { JSON_READER_ERROR_INVALID_TYPE, "JSON_READER_ERROR_INVALID_TYPE", "invalid-type" },
        { 0, NULL, NULL }
      };

      GType g_enum_type_id =
        g_enum_register_static (g_intern_static_string ("JsonReaderError"), values);

      g_once_init_leave (&g_enum_type_id__volatile, g_enum_type_id);
    }

  return g_enum_type_id__volatile;
}

/* enumerations from "json-types.h" */
#include "json-types.h"
GType
json_node_type_get_type (void)
{
  static gsize g_enum_type_id__volatile = 0;

  if (g_once_init_enter (&g_enum_type_id__volatile))
    {
      static const GEnumValue values[] = {
        { JSON_NODE_OBJECT, "JSON_NODE_OBJECT", "object" },
        { JSON_NODE_ARRAY, "JSON_NODE_ARRAY", "array" },
        { JSON_NODE_VALUE, "JSON_NODE_VALUE", "value" },
        { JSON_NODE_NULL, "JSON_NODE_NULL", "null" },
        { 0, NULL, NULL }
      };

      GType g_enum_type_id =
        g_enum_register_static (g_intern_static_string ("JsonNodeType"), values);

      g_once_init_leave (&g_enum_type_id__volatile, g_enum_type_id);
    }

  return g_enum_type_id__volatile;
}

/* Generated data ends here */

