/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN:OBJECT,STRING,VARIANT,VARIANT (../src/jsonrpc-marshalers.list:1) */
G_GNUC_INTERNAL
void _jsonrpc_marshal_BOOLEAN__OBJECT_STRING_VARIANT_VARIANT (GClosure     *closure,
                                                              GValue       *return_value,
                                                              guint         n_param_values,
                                                              const GValue *param_values,
                                                              gpointer      invocation_hint,
                                                              gpointer      marshal_data);
G_GNUC_INTERNAL
void _jsonrpc_marshal_BOOLEAN__OBJECT_STRING_VARIANT_VARIANTv (GClosure *closure,
                                                               GValue   *return_value,
                                                               gpointer  instance,
                                                               va_list   args,
                                                               gpointer  marshal_data,
                                                               int       n_params,
                                                               GType    *param_types);

/* BOOLEAN:STRING,VARIANT,VARIANT (../src/jsonrpc-marshalers.list:2) */
G_GNUC_INTERNAL
void _jsonrpc_marshal_BOOLEAN__STRING_VARIANT_VARIANT (GClosure     *closure,
                                                       GValue       *return_value,
                                                       guint         n_param_values,
                                                       const GValue *param_values,
                                                       gpointer      invocation_hint,
                                                       gpointer      marshal_data);
G_GNUC_INTERNAL
void _jsonrpc_marshal_BOOLEAN__STRING_VARIANT_VARIANTv (GClosure *closure,
                                                        GValue   *return_value,
                                                        gpointer  instance,
                                                        va_list   args,
                                                        gpointer  marshal_data,
                                                        int       n_params,
                                                        GType    *param_types);

/* VOID:OBJECT,STRING,VARIANT (../src/jsonrpc-marshalers.list:3) */
G_GNUC_INTERNAL
void _jsonrpc_marshal_VOID__OBJECT_STRING_VARIANT (GClosure     *closure,
                                                   GValue       *return_value,
                                                   guint         n_param_values,
                                                   const GValue *param_values,
                                                   gpointer      invocation_hint,
                                                   gpointer      marshal_data);
G_GNUC_INTERNAL
void _jsonrpc_marshal_VOID__OBJECT_STRING_VARIANTv (GClosure *closure,
                                                    GValue   *return_value,
                                                    gpointer  instance,
                                                    va_list   args,
                                                    gpointer  marshal_data,
                                                    int       n_params,
                                                    GType    *param_types);

/* VOID:STRING,VARIANT (../src/jsonrpc-marshalers.list:4) */
G_GNUC_INTERNAL
void _jsonrpc_marshal_VOID__STRING_VARIANT (GClosure     *closure,
                                            GValue       *return_value,
                                            guint         n_param_values,
                                            const GValue *param_values,
                                            gpointer      invocation_hint,
                                            gpointer      marshal_data);
G_GNUC_INTERNAL
void _jsonrpc_marshal_VOID__STRING_VARIANTv (GClosure *closure,
                                             GValue   *return_value,
                                             gpointer  instance,
                                             va_list   args,
                                             gpointer  marshal_data,
                                             int       n_params,
                                             GType    *param_types);


G_END_DECLS
